"""
A module implementing the package's runtime environment.
"""

# built-in
from contextlib import ExitStack, contextmanager
from typing import Iterator

# third-party
from vcorelib.paths import Pathlike

# internal
from rcmpy.environment.template import TemplateEnvironment
from rcmpy.state import load_state


class Environment(TemplateEnvironment):
    """A class implementing this package's runtime environment."""


@contextmanager
def load_environment(
    root: Pathlike = None, name: str = "state.json"
) -> Iterator[Environment]:
    """A wrapper for loading an environment with default state data."""

    with ExitStack() as stack:
        yield Environment(
            stack.enter_context(load_state(root=root, name=name)), stack
        )



"""
An environment extension for working with template files.
"""

# third-party
from datazen.templates import environment
from jinja2 import FileSystemLoader

# internal
from rcmpy.environment.base import BaseEnvironment


class TemplateEnvironment(BaseEnvironment):
    """
    A class implementing template management for this package's runtime
    environment.
    """

    def _init_loaded(self) -> None:
        """Called during initialization if a valid configuration is loaded."""

        super()._init_loaded()

        if not hasattr(self, "jinja"):
            template_root = self.state.directory.joinpath("templates")

            template_dirs = [
                x
                for x in [
                    template_root.joinpath(self.state.variant),
                    template_root.joinpath("common"),
                ]
                if x.is_dir()
            ]

            # Prefer variant templates, if the variant template-directory
            # exists.
            self.jinja = environment(
                loader=FileSystemLoader(
                    template_dirs,
                    followlinks=True,
                )
            )

            # Tasks to do:
            # - split into two kinds of files, literal files and actual jinja
            #   templates (check for .j2 extension)
            # - find the full, absolute paths of the template objects
            # - add the template dirs to the template file-info cache
            # - use vcorelib file-info cache to figure out which templates are
            #   new or changed (log this?)
            for template in self.jinja.list_templates():
                print(template)



"""
A module implementing a basic environment interface.
"""

# built-in
from contextlib import ExitStack
from typing import Optional

# third-party
from vcorelib.io.types import FileExtension
from vcorelib.logging import LoggerMixin

# internal
from rcmpy import PKG_NAME
from rcmpy.config import Config
from rcmpy.state import State


class BaseEnvironment(LoggerMixin):
    """A class implementing this package's base runtime environment."""

    def __init__(self, state: State, stack: ExitStack) -> None:
        """Initialize this instance."""

        super().__init__()
        self.state = state
        self.stack = stack
        self._config: Optional[Config] = None

        config_base = state.directory.joinpath(PKG_NAME)

        config_candidates = list(
            FileExtension.data_candidates(config_base, exists_only=True)
        )

        if not config_candidates:
            self.logger.error(
                "No config found: %s.",
                list(
                    str(x) for x in FileExtension.data_candidates(config_base)
                ),
            )
        else:
            assert len(config_candidates) == 1, config_candidates
            self._config = Config.decode(config_candidates[0])
            self.logger.info("Loaded config '%s'.", config_candidates[0])
            self._init_loaded()

    def _init_loaded(self) -> None:
        """Called during initialization if a valid configuration is loaded."""

    @property
    def config_loaded(self) -> bool:
        """Determine if this environment has loaded a config."""
        return self._config is not None

    @property
    def config(self) -> Config:
        """Get this environment's configuration object."""
        assert self._config is not None, "Config not loaded!"
        return self._config



"""
A module implementing this package's stateful information that persists across
invocations.
"""

# built-in
from contextlib import contextmanager
from logging import getLogger
from pathlib import Path
from typing import Iterator, cast

# third-party
from vcorelib.dict.cache import FileCache
from vcorelib.io.types import JsonObject as _JsonObject
from vcorelib.paths import Pathlike, normalize

# internal
from rcmpy.paths import default_config_directory, default_state_directory
from rcmpy.schemas import RcmpyDictCodec as _RcmpyDictCodec

LOG = getLogger(__name__)


class State(_RcmpyDictCodec):
    """The top-level configuration object for the package."""

    directory: Path
    variant: str

    def init(self, data: _JsonObject) -> None:
        """Perform implementation-specific initialization."""

        self.logger = LOG

        self.directory = normalize(
            cast(str, data.get("directory", default_config_directory()))
        ).resolve()

        self.logger.info("Using directory '%s'.", self.directory)

        self.variant: str = cast(str, data["variant"])
        if self.variant:
            self.logger.info("Using variant '%s'.", self.variant)

    def set_directory(self, path: Pathlike) -> None:
        """Set a new directory to use as the data repository."""

        new_dir = normalize(path).resolve()

        if new_dir == self.directory:
            self.logger.info("New directory '%s' same as current.", new_dir)
            return

        self.directory = new_dir
        self.logger.info("Set directory to '%s'.", new_dir)

    def set_variant(self, variant: str = None) -> None:
        """Set a new variant value."""

        if not variant:
            self.logger.info("Current variant: '%s'.", self.variant)
        elif variant != self.variant:
            self.variant = variant
            self.logger.info("Updating variant to '%s'.", variant)

    def asdict(self) -> _JsonObject:
        """Obtain a dictionary representing this instance."""

        return {"directory": str(self.directory), "variant": self.variant}


@contextmanager
def load_state(
    root: Pathlike = None, name: str = "state.json"
) -> Iterator[State]:
    """This needs to be a context manager, so it get's written back to disk."""

    if root is None:
        root = default_state_directory()

    with FileCache(normalize(root, name)).loaded() as data:
        state = State.create(data, verify=False)
        yield state

        # Update the original dictionary with any changes to the state object.
        data.update(state.asdict())



"""
A module for common argument-parsing methods.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser


def add_default_flag(parser: _ArgumentParser) -> None:
    """Adds a default-flag argument."""

    parser.add_argument(
        "-d",
        "--default",
        action="store_true",
        help="sets the directory back to the package default",
    )



# =====================================
# generator=datazen
# version=3.1.2
# hash=e990bc9fe3294af3d40349e7b0627ad4
# =====================================

"""
A module aggregating package commands.
"""

# built-in
from typing import List as _List
from typing import Tuple as _Tuple

# third-party
from vcorelib.args import CommandRegister as _CommandRegister

# internal
from rcmpy.commands.apply import add_apply_cmd
from rcmpy.commands.use import add_use_cmd
from rcmpy.commands.variant import add_variant_cmd
from rcmpy.commands.watch import add_watch_cmd


def commands() -> _List[_Tuple[str, str, _CommandRegister]]:
    """Get this package's commands."""

    return [
        (
            "apply",
            "apply any pending changes from the active data repository",
            add_apply_cmd,
        ),
        (
            "use",
            "set the directory to use as the rcmpy data repository",
            add_use_cmd,
        ),
        (
            "variant",
            "set the variant of configuration data to use",
            add_variant_cmd,
        ),
        (
            "watch",
            "do a task whenever a file in a specified directory changes",
            add_watch_cmd,
        ),
        ("noop", "command stub (does nothing)", lambda _: lambda _: 0),
    ]



"""
An entry-point for the 'variant' command.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace

# third-party
from vcorelib.args import CommandFunction as _CommandFunction

# internal
from rcmpy.commands.common import add_default_flag
from rcmpy.state import load_state


def variant_cmd(args: _Namespace) -> int:
    """Execute the variant command."""

    with load_state() as state:
        if args.default:
            args.variant = "default"

        state.set_variant(args.variant)

    return 0


def add_variant_cmd(parser: _ArgumentParser) -> _CommandFunction:
    """Add variant-command arguments to its parser."""

    add_default_flag(parser)

    parser.add_argument("variant", nargs="?", help="new variant to use")
    return variant_cmd



"""
A module for working with the XDG base directory specification.

https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
"""

# built-in
from os import environ
from os import path as _path
from pathlib import Path
from typing import Iterator

# third-party
from vcorelib.paths import Pathlike, normalize


def ensure_absolute(path: Pathlike, create: bool = False) -> Path:
    """A wrapper to ensure that a path is an absolute one."""

    path = normalize(path)
    assert path.is_absolute(), f"'{path}' isn't absolute!"

    if create:
        path.mkdir(mode=0o700, parents=True, exist_ok=True)

    return path


def user_data(*parts: str, create: bool = True) -> Path:
    """
    A single base directory relative to which user-specific data files should
    be written.
    """

    return ensure_absolute(
        Path(
            environ.get(
                "XDG_DATA_HOME", Path.home().joinpath(".local", "share")
            ),
            *parts,
        ),
        create=create,
    )


def user_config(*parts: str, create: bool = True) -> Path:
    """
    A single base directory relative to which user-specific configuration files
    should be written.
    """

    return ensure_absolute(
        Path(
            environ.get("XDG_CONFIG_HOME", Path.home().joinpath(".config")),
            *parts,
        ),
        create=create,
    )


def user_state(*parts: str, create: bool = True) -> Path:
    """
    A single base directory relative to which user-specific state data
    should be written.
    """

    return ensure_absolute(
        Path(
            environ.get(
                "XDG_STATE_HOME", Path.home().joinpath(".local", "state")
            ),
            *parts,
        ),
        create=create,
    )


def user_bin(*parts: str, create: bool = True) -> Path:
    """
    A single base directory relative to which user-specific executable files
    may be written.
    """

    return ensure_absolute(
        Path.home().joinpath(".local", "bin", *parts), create=create
    )


def iterate_directories(
    data: str, sep: str = _path.pathsep, create: bool = False
) -> Iterator[Path]:
    """
    Iterate over directories specified in a PATH-like variable (some
    separator).
    """

    for item in data.split(sep):
        yield ensure_absolute(Path(item), create=create)


def root_directory(*parts: str) -> Path:
    """Create a path from the current file-system's root directory."""
    return Path(_path.abspath(".").split(_path.sep)[0] + _path.sep, *parts)


def data_directories(*parts: str, include_home: bool = True) -> Iterator[Path]:
    """
    Yield a set of preference ordered base directories relative to which
    data files should be searched.
    """

    if include_home:
        yield user_data(*parts)

    yield from iterate_directories(
        environ.get(
            "XDG_DATA_DIRS",
            _path.pathsep.join(
                [
                    str(root_directory("share", *parts)),
                    str(root_directory("local", "share", *parts)),
                ]
            ),
        )
    )


def config_directories(
    *parts: str, include_home: bool = True
) -> Iterator[Path]:
    """
    Yield a set of preference ordered base directories relative to which
    configuration files should be searched.
    """

    if include_home:
        yield user_config(*parts)

    yield from iterate_directories(
        environ.get(
            "XDG_CONFIG_DIRS", str(root_directory("etc", "xdg", *parts))
        )
    )


def user_cache(*parts: str, create: bool = True) -> Path:
    """
    A single base directory relative to which user-specific
    non-essential (cached) data should be written.
    """

    return ensure_absolute(
        Path(
            environ.get("XDG_CACHE_HOME", Path.home().joinpath(".cache")),
            *parts,
        ),
        create=create,
    )



"""
A module implementing a configuration interface for the package.
"""

# third-party
from vcorelib.dict.codec import BasicDictCodec as _BasicDictCodec

# internal
from rcmpy.schemas import RcmpyDictCodec as _RcmpyDictCodec


class Config(_RcmpyDictCodec, _BasicDictCodec):
    """The top-level configuration object for the package."""



"""
An entry-point for the 'apply' command.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace

# third-party
from vcorelib.args import CommandFunction as _CommandFunction

# internal
from rcmpy.environment import load_environment


def apply_cmd(_: _Namespace) -> int:
    """Execute the apply command."""

    with load_environment() as env:
        if not env.config_loaded:
            return 1

        print(env.config)

    return 0


def add_apply_cmd(_: _ArgumentParser) -> _CommandFunction:
    """Add apply-command arguments to its parser."""

    return apply_cmd



# =====================================
# generator=datazen
# version=3.1.2
# hash=5f72b7e25844456d9d2f200300b91b3e
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "A configuration-file management system."
PKG_NAME = "rcmpy"
VERSION = "0.2.0"



# =====================================
# generator=datazen
# version=3.1.2
# hash=dc0d9f835ecbdaf7ba7d62dc617acb8d
# =====================================

"""
rcmpy - Package definition for distribution.
"""

# third-party
try:
    from setuptools_wrapper.setup import setup
except (ImportError, ModuleNotFoundError):
    from rcmpy_bootstrap.setup import setup  # type: ignore

# internal
from rcmpy import DESCRIPTION, PKG_NAME, VERSION

author_info = {
    "name": "Vaughn Kottler",
    "email": "vaughnkottler@gmail.com",
    "username": "vkottler",
}
pkg_info = {
    "name": PKG_NAME,
    "slug": PKG_NAME.replace("-", "_"),
    "version": VERSION,
    "description": DESCRIPTION,
    "versions": [
        "3.7",
        "3.8",
        "3.9",
        "3.10",
        "3.11",
    ],
}
setup(
    pkg_info,
    author_info,
)



"""
An entry-point for the 'watch' command.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace
from asyncio import Event, get_event_loop, sleep
from logging import getLogger
from os import getpid
from pathlib import Path

# third-party
from vcorelib.args import CommandFunction as _CommandFunction
from vcorelib.asyncio import run_handle_stop
from vcorelib.asyncio.cli import run_command, run_shell
from vcorelib.paths.info_cache import FileChanged, file_info_cache

# internal
from rcmpy.paths import default_cache_directory

LOG = getLogger(__name__)


def watch_cache() -> Path:
    """Obtain a file path to use as a cache."""
    return default_cache_directory().joinpath(f"watch_cache-{getpid()}.json")


async def command(*args: str, shell: bool = False) -> int:
    """Run a subprocess and return the return code."""

    runner = run_shell if shell else run_command
    proc = await runner(LOG, *args)
    assert proc.returncode is not None
    return proc.returncode


async def entry(stop_sig: Event, args: _Namespace) -> int:
    """The async entry-point for the watch command."""

    cache_file = watch_cache()

    count = 0

    def poll_cb(_: FileChanged) -> bool:
        """Method to run when any watched file changes."""

        nonlocal count
        count += 1
        return True

    with file_info_cache(cache_file, poll_cb, logger=LOG) as files:
        while not stop_sig.is_set():
            files.poll_directory(args.directory, base=args.dir)
            files.poll_existing(base=args.dir)

            if count > 0:
                # Run the command, return True if it exits 0.
                await command(*args.cmd, shell=args.shell)
                count = 0

            if args.single_pass:
                stop_sig.set()

            # Would be neat to implement something to keep this going at
            # a regular period
            await sleep(args.poll_rate)

    # The cache file shouldn't persist across invocations.
    cache_file.unlink()
    return 0


def watch_cmd(args: _Namespace) -> int:
    """Execute the watch command."""

    eloop = get_event_loop()
    stop_sig = Event()
    return run_handle_stop(stop_sig, entry(stop_sig, args), eloop=eloop)


def add_watch_cmd(parser: _ArgumentParser) -> _CommandFunction:
    """Add watch-command arguments to its parser."""

    parser.add_argument(
        "-p",
        "--poll-rate",
        type=float,
        default=0.1,
        help="poll period in seconds (default: %(default)ss)",
    )
    parser.add_argument(
        "-s", "--shell", action="store_true", help="set to run a shell command"
    )
    parser.add_argument(
        "-i",
        "--single-pass",
        action="store_true",
        help="only run a single iteration",
    )
    parser.add_argument(
        "directory", type=Path, help="directory to watch for file changes"
    )
    parser.add_argument("cmd", nargs="+", help="command to run")

    return watch_cmd



"""
Test the 'xdg' module.
"""

# module under test
from rcmpy import xdg


def test_xdg_basic():
    """Test basic XDG interfaces."""

    assert xdg.user_data()
    assert xdg.user_config()
    assert xdg.user_state()
    assert xdg.user_bin()

    assert list(xdg.data_directories())
    assert list(xdg.config_directories())

    assert xdg.user_cache()
